@@ -17,15 +17,17 @@
  * Boston, MA 021110-1307, USA.
  */
 
-#ifndef __CHUNK_CHECK_H__
-#define __CHUNK_CHECK_H__
+#ifndef __BTRFS_CHECK_H__
+#define __BTRFS_CHECK_H__
 
 #if BTRFS_FLAT_INCLUDES
 #include "kerncompat.h"
+#include "ctree.h"
 #include "extent-cache.h"
 #include "list.h"
 #else
 #include <btrfs/kerncompat.h>
+#include <btrfs/ctree.h>
 #include <btrfs/extent-cache.h>
 #include <btrfs/list.h>
 #endif /* BTRFS_FLAT_INCLUDES */
@@ -179,5 +181,6 @@ btrfs_new_device_extent_record(struct extent_buffer *leaf,
 int check_chunks(struct cache_tree *chunk_cache,
 		 struct block_group_tree *block_group_cache,
 		 struct device_extent_tree *dev_extent_cache,
-		 struct list_head *good, struct list_head *bad, int silent);
+		 struct list_head *good, struct list_head *bad,
+		 struct list_head *rebuild, int silent);
 #endif
