@@ -56,18 +56,32 @@ struct btrfs_receive
 	int dest_dir_fd;
 
 	int write_fd;
-	char *write_path;
+	char write_path[PATH_MAX];
 
 	char *root_path;
 	char *dest_dir_path; /* relative to root_path */
-	char *full_subvol_path;
+	char full_subvol_path[PATH_MAX];
+	char *full_root_path;
 	int dest_dir_chroot;
 
-	struct subvol_info *cur_subvol;
+	struct subvol_info cur_subvol;
+	/*
+	 * Substitute for cur_subvol::path which is a pointer and we cannot
+	 * change it to an array as it's a public API.
+	 */
+	char cur_subvol_path[PATH_MAX];
 
 	struct subvol_uuid_search sus;
 
 	int honor_end_cmd;
+
+	/*
+	 * Buffer to store capabilities from security.capabilities xattr,
+	 * usually 20 bytes, but make same room for potentially larger
+	 * encodings. Must be set only once per file, denoted by length > 0.
+	 */
+	char cached_capabilities[64];
+	int cached_capabilities_len;
 };
 
 static int finish_subvol(struct btrfs_receive *r)
@@ -78,21 +92,21 @@ static int finish_subvol(struct btrfs_receive *r)
 	char uuid_str[BTRFS_UUID_UNPARSED_SIZE];
 	u64 flags;
 
-	if (r->cur_subvol == NULL)
+	if (r->cur_subvol_path[0] == 0)
 		return 0;
 
-	subvol_fd = openat(r->mnt_fd, r->cur_subvol->path,
+	subvol_fd = openat(r->mnt_fd, r->cur_subvol_path,
 			O_RDONLY | O_NOATIME);
 	if (subvol_fd < 0) {
 		ret = -errno;
 		fprintf(stderr, "ERROR: open %s failed. %s\n",
-				r->cur_subvol->path, strerror(-ret));
+				r->cur_subvol_path, strerror(-ret));
 		goto out;
 	}
 
 	memset(&rs_args, 0, sizeof(rs_args));
-	memcpy(rs_args.uuid, r->cur_subvol->received_uuid, BTRFS_UUID_SIZE);
-	rs_args.stransid = r->cur_subvol->stransid;
+	memcpy(rs_args.uuid, r->cur_subvol.received_uuid, BTRFS_UUID_SIZE);
+	rs_args.stransid = r->cur_subvol.stransid;
 
 	if (g_verbose >= 1) {
 		uuid_unparse((u8*)rs_args.uuid, uuid_str);
@@ -107,7 +121,7 @@ static int finish_subvol(struct btrfs_receive *r)
 				strerror(-ret));
 		goto out;
 	}
-	r->cur_subvol->rtransid = rs_args.rtransid;
+	r->cur_subvol.rtransid = rs_args.rtransid;
 
 	ret = ioctl(subvol_fd, BTRFS_IOC_SUBVOL_GETFLAGS, &flags);
 	if (ret < 0) {
@@ -130,10 +144,8 @@ static int finish_subvol(struct btrfs_receive *r)
 	ret = 0;
 
 out:
-	if (r->cur_subvol) {
-		free(r->cur_subvol->path);
-		free(r->cur_subvol);
-		r->cur_subvol = NULL;
+	if (r->cur_subvol_path[0]) {
+		r->cur_subvol_path[0] = 0;
 	}
 	if (subvol_fd != -1)
 		close(subvol_fd);
@@ -152,25 +164,36 @@ static int process_subvol(const char *path, const u8 *uuid, u64 ctransid,
 	if (ret < 0)
 		goto out;
 
-	r->cur_subvol = calloc(1, sizeof(*r->cur_subvol));
+	BUG_ON(r->cur_subvol.path);
+	BUG_ON(r->cur_subvol_path[0]);
 
-	if (strlen(r->dest_dir_path) == 0)
-		r->cur_subvol->path = strdup(path);
-	else
-		r->cur_subvol->path = path_cat(r->dest_dir_path, path);
-	free(r->full_subvol_path);
-	r->full_subvol_path = path_cat3(r->root_path, r->dest_dir_path, path);
+	if (strlen(r->dest_dir_path) == 0) {
+		strncpy_null(r->cur_subvol_path, path);
+	} else {
+		ret = path_cat_out(r->cur_subvol_path, r->dest_dir_path, path);
+		if (ret < 0) {
+			fprintf(stderr, "ERROR: subvol: path invalid: %s\n",
+					path);
+			goto out;
+		}
+	}
+	ret = path_cat3_out(r->full_subvol_path, r->root_path,
+			r->dest_dir_path, path);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: subvol: path invalid: %s\n", path);
+		goto out;
+	}
 
 	fprintf(stderr, "At subvol %s\n", path);
 
-	memcpy(r->cur_subvol->received_uuid, uuid, BTRFS_UUID_SIZE);
-	r->cur_subvol->stransid = ctransid;
+	memcpy(r->cur_subvol.received_uuid, uuid, BTRFS_UUID_SIZE);
+	r->cur_subvol.stransid = ctransid;
 
 	if (g_verbose) {
-		uuid_unparse((u8*)r->cur_subvol->received_uuid, uuid_str);
+		uuid_unparse((u8*)r->cur_subvol.received_uuid, uuid_str);
 		fprintf(stderr, "receiving subvol %s uuid=%s, stransid=%llu\n",
 				path, uuid_str,
-				r->cur_subvol->stransid);
+				r->cur_subvol.stransid);
 	}
 
 	memset(&args_v1, 0, sizeof(args_v1));
@@ -201,25 +224,36 @@ static int process_snapshot(const char *path, const u8 *uuid, u64 ctransid,
 	if (ret < 0)
 		goto out;
 
-	r->cur_subvol = calloc(1, sizeof(*r->cur_subvol));
+	BUG_ON(r->cur_subvol.path);
+	BUG_ON(r->cur_subvol_path[0]);
 
-	if (strlen(r->dest_dir_path) == 0)
-		r->cur_subvol->path = strdup(path);
-	else
-		r->cur_subvol->path = path_cat(r->dest_dir_path, path);
-	free(r->full_subvol_path);
-	r->full_subvol_path = path_cat3(r->root_path, r->dest_dir_path, path);
+	if (strlen(r->dest_dir_path) == 0) {
+		strncpy_null(r->cur_subvol_path, path);
+	} else {
+		ret = path_cat_out(r->cur_subvol_path, r->dest_dir_path, path);
+		if (ret < 0) {
+			fprintf(stderr, "ERROR: snapshot: path invalid: %s\n",
+					path);
+			goto out;
+		}
+	}
+	ret = path_cat3_out(r->full_subvol_path, r->root_path,
+			r->dest_dir_path, path);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: snapshot: path invalid: %s\n", path);
+		goto out;
+	}
 
 	fprintf(stdout, "At snapshot %s\n", path);
 
-	memcpy(r->cur_subvol->received_uuid, uuid, BTRFS_UUID_SIZE);
-	r->cur_subvol->stransid = ctransid;
+	memcpy(r->cur_subvol.received_uuid, uuid, BTRFS_UUID_SIZE);
+	r->cur_subvol.stransid = ctransid;
 
 	if (g_verbose) {
-		uuid_unparse((u8*)r->cur_subvol->received_uuid, uuid_str);
+		uuid_unparse((u8*)r->cur_subvol.received_uuid, uuid_str);
 		fprintf(stderr, "receiving snapshot %s uuid=%s, "
 				"ctransid=%llu ", path, uuid_str,
-				r->cur_subvol->stransid);
+				r->cur_subvol.stransid);
 		uuid_unparse(parent_uuid, uuid_str);
 		fprintf(stderr, "parent_uuid=%s, parent_ctransid=%llu\n",
 				uuid_str, parent_ctransid);
@@ -240,6 +274,46 @@ static int process_snapshot(const char *path, const u8 *uuid, u64 ctransid,
 		goto out;
 	}
 
+	/*
+	 * The path is resolved from the root subvol, but we could be in some
+	 * subvolume under the root subvolume, so try and adjust the path to be
+	 * relative to our root path.
+	 */
+	if (r->full_root_path) {
+		size_t root_len;
+		size_t sub_len;
+
+		root_len = strlen(r->full_root_path);
+		sub_len = strlen(parent_subvol->path);
+
+		/* First make sure the parent subvol is actually in our path */
+		if (sub_len < root_len ||
+		    strstr(parent_subvol->path, r->full_root_path) == NULL) {
+			fprintf(stderr, "ERROR: parent subvol is not reachable"
+				" from inside the root subvol.\n");
+			ret = -ENOENT;
+			goto out;
+		}
+
+		if (sub_len == root_len) {
+			parent_subvol->path[0] = '/';
+			parent_subvol->path[1] = '\0';
+		} else {
+			/*
+			 * root path is foo/bar
+			 * subvol path is foo/bar/baz
+			 *
+			 * we need to have baz be the path, so we need to move
+			 * the bit after foo/bar/, so path + root_len + 1, and
+			 * move the part we care about, so sub_len - root_len -
+			 * 1.
+			 */
+			memmove(parent_subvol->path,
+				parent_subvol->path + root_len + 1,
+				sub_len - root_len - 1);
+			parent_subvol->path[sub_len - root_len - 1] = '\0';
+		}
+	}
 	/*if (rs_args.ctransid > rs_args.rtransid) {
 		if (!r->force) {
 			ret = -EINVAL;
@@ -250,8 +324,11 @@ static int process_snapshot(const char *path, const u8 *uuid, u64 ctransid,
 		}
 	}*/
 
-	args_v2.fd = openat(r->mnt_fd, parent_subvol->path,
-			O_RDONLY | O_NOATIME);
+	if (strlen(parent_subvol->path) == 0)
+		args_v2.fd = dup(r->mnt_fd);
+	else
+		args_v2.fd = openat(r->mnt_fd, parent_subvol->path,
+				O_RDONLY | O_NOATIME);
 	if (args_v2.fd < 0) {
 		ret = -errno;
 		if (errno != ENOENT)
@@ -288,7 +365,13 @@ static int process_mkfile(const char *path, void *user)
 {
 	int ret;
 	struct btrfs_receive *r = user;
-	char *full_path = path_cat(r->full_subvol_path, path);
+	char full_path[PATH_MAX];
+
+	ret = path_cat_out(full_path, r->full_subvol_path, path);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: mkfile: path invalid: %s\n", path);
+		goto out;
+	}
 
 	if (g_verbose >= 2)
 		fprintf(stderr, "mkfile %s\n", path);
@@ -304,7 +387,6 @@ static int process_mkfile(const char *path, void *user)
 	ret = 0;
 
 out:
-	free(full_path);
 	return ret;
 }
 
@@ -312,7 +394,13 @@ static int process_mkdir(const char *path, void *user)
 {
 	int ret;
 	struct btrfs_receive *r = user;
-	char *full_path = path_cat(r->full_subvol_path, path);
+	char full_path[PATH_MAX];
+
+	ret = path_cat_out(full_path, r->full_subvol_path, path);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: mkdir: path invalid: %s\n", path);
+		goto out;
+	}
 
 	if (g_verbose >= 2)
 		fprintf(stderr, "mkdir %s\n", path);
@@ -324,7 +412,7 @@ static int process_mkdir(const char *path, void *user)
 				strerror(-ret));
 	}
 
-	free(full_path);
+out:
 	return ret;
 }
 
@@ -332,7 +420,13 @@ static int process_mknod(const char *path, u64 mode, u64 dev, void *user)
 {
 	int ret;
 	struct btrfs_receive *r = user;
-	char *full_path = path_cat(r->full_subvol_path, path);
+	char full_path[PATH_MAX];
+
+	ret = path_cat_out(full_path, r->full_subvol_path, path);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: mknod: path invalid: %s\n", path);
+		goto out;
+	}
 
 	if (g_verbose >= 2)
 		fprintf(stderr, "mknod %s mode=%llu, dev=%llu\n",
@@ -345,7 +439,7 @@ static int process_mknod(const char *path, u64 mode, u64 dev, void *user)
 				strerror(-ret));
 	}
 
-	free(full_path);
+out:
 	return ret;
 }
 
@@ -353,7 +447,13 @@ static int process_mkfifo(const char *path, void *user)
 {
 	int ret;
 	struct btrfs_receive *r = user;
-	char *full_path = path_cat(r->full_subvol_path, path);
+	char full_path[PATH_MAX];
+
+	ret = path_cat_out(full_path, r->full_subvol_path, path);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: mkfifo: path invalid: %s\n", path);
+		goto out;
+	}
 
 	if (g_verbose >= 2)
 		fprintf(stderr, "mkfifo %s\n", path);
@@ -365,7 +465,7 @@ static int process_mkfifo(const char *path, void *user)
 				strerror(-ret));
 	}
 
-	free(full_path);
+out:
 	return ret;
 }
 
@@ -373,7 +473,13 @@ static int process_mksock(const char *path, void *user)
 {
 	int ret;
 	struct btrfs_receive *r = user;
-	char *full_path = path_cat(r->full_subvol_path, path);
+	char full_path[PATH_MAX];
+
+	ret = path_cat_out(full_path, r->full_subvol_path, path);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: mksock: path invalid: %s\n", path);
+		goto out;
+	}
 
 	if (g_verbose >= 2)
 		fprintf(stderr, "mksock %s\n", path);
@@ -385,7 +491,7 @@ static int process_mksock(const char *path, void *user)
 				strerror(-ret));
 	}
 
-	free(full_path);
+out:
 	return ret;
 }
 
@@ -393,7 +499,13 @@ static int process_symlink(const char *path, const char *lnk, void *user)
 {
 	int ret;
 	struct btrfs_receive *r = user;
-	char *full_path = path_cat(r->full_subvol_path, path);
+	char full_path[PATH_MAX];
+
+	ret = path_cat_out(full_path, r->full_subvol_path, path);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: symlink: path invalid: %s\n", path);
+		goto out;
+	}
 
 	if (g_verbose >= 2)
 		fprintf(stderr, "symlink %s -> %s\n", path, lnk);
@@ -405,7 +517,7 @@ static int process_symlink(const char *path, const char *lnk, void *user)
 				lnk, strerror(-ret));
 	}
 
-	free(full_path);
+out:
 	return ret;
 }
 
@@ -413,8 +525,22 @@ static int process_rename(const char *from, const char *to, void *user)
 {
 	int ret;
 	struct btrfs_receive *r = user;
-	char *full_from = path_cat(r->full_subvol_path, from);
-	char *full_to = path_cat(r->full_subvol_path, to);
+	char full_from[PATH_MAX];
+	char full_to[PATH_MAX];
+
+	ret = path_cat_out(full_from, r->full_subvol_path, from);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: rename: source path invalid: %s\n",
+				from);
+		goto out;
+	}
+
+	ret = path_cat_out(full_to, r->full_subvol_path, to);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: rename: target path invalid: %s\n",
+				to);
+		goto out;
+	}
 
 	if (g_verbose >= 2)
 		fprintf(stderr, "rename %s -> %s\n", from, to);
@@ -426,8 +552,7 @@ static int process_rename(const char *from, const char *to, void *user)
 				to, strerror(-ret));
 	}
 
-	free(full_from);
-	free(full_to);
+out:
 	return ret;
 }
 
@@ -435,8 +560,22 @@ static int process_link(const char *path, const char *lnk, void *user)
 {
 	int ret;
 	struct btrfs_receive *r = user;
-	char *full_path = path_cat(r->full_subvol_path, path);
-	char *full_link_path = path_cat(r->full_subvol_path, lnk);
+	char full_path[PATH_MAX];
+	char full_link_path[PATH_MAX];
+
+	ret = path_cat_out(full_path, r->full_subvol_path, path);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: link: source path invalid: %s\n",
+				full_path);
+		goto out;
+	}
+
+	ret = path_cat_out(full_link_path, r->full_subvol_path, lnk);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: link: target path invalid: %s\n",
+				full_link_path);
+		goto out;
+	}
 
 	if (g_verbose >= 2)
 		fprintf(stderr, "link %s -> %s\n", path, lnk);
@@ -448,8 +587,7 @@ static int process_link(const char *path, const char *lnk, void *user)
 				lnk, strerror(-ret));
 	}
 
-	free(full_path);
-	free(full_link_path);
+out:
 	return ret;
 }
 
@@ -458,7 +596,13 @@ static int process_unlink(const char *path, void *user)
 {
 	int ret;
 	struct btrfs_receive *r = user;
-	char *full_path = path_cat(r->full_subvol_path, path);
+	char full_path[PATH_MAX];
+
+	ret = path_cat_out(full_path, r->full_subvol_path, path);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: unlink: path invalid: %s\n", path);
+		goto out;
+	}
 
 	if (g_verbose >= 2)
 		fprintf(stderr, "unlink %s\n", path);
@@ -470,7 +614,7 @@ static int process_unlink(const char *path, void *user)
 				strerror(-ret));
 	}
 
-	free(full_path);
+out:
 	return ret;
 }
 
@@ -478,7 +622,13 @@ static int process_rmdir(const char *path, void *user)
 {
 	int ret;
 	struct btrfs_receive *r = user;
-	char *full_path = path_cat(r->full_subvol_path, path);
+	char full_path[PATH_MAX];
+
+	ret = path_cat_out(full_path, r->full_subvol_path, path);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: rmdir: path invalid: %s\n", path);
+		goto out;
+	}
 
 	if (g_verbose >= 2)
 		fprintf(stderr, "rmdir %s\n", path);
@@ -490,11 +640,10 @@ static int process_rmdir(const char *path, void *user)
 				strerror(-ret));
 	}
 
-	free(full_path);
+out:
 	return ret;
 }
 
-
 static int open_inode_for_write(struct btrfs_receive *r, const char *path)
 {
 	int ret = 0;
@@ -513,8 +662,7 @@ static int open_inode_for_write(struct btrfs_receive *r, const char *path)
 				strerror(-ret));
 		goto out;
 	}
-	free(r->write_path);
-	r->write_path = strdup(path);
+	strncpy_null(r->write_path, path);
 
 out:
 	return ret;
@@ -535,10 +683,16 @@ static int process_write(const char *path, const void *data, u64 offset,
 {
 	int ret = 0;
 	struct btrfs_receive *r = user;
-	char *full_path = path_cat(r->full_subvol_path, path);
+	char full_path[PATH_MAX];
 	u64 pos = 0;
 	int w;
 
+	ret = path_cat_out(full_path, r->full_subvol_path, path);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: write: path invalid: %s\n", path);
+		goto out;
+	}
+
 	ret = open_inode_for_write(r, full_path);
 	if (ret < 0)
 		goto out;
@@ -556,7 +710,6 @@ static int process_write(const char *path, const void *data, u64 offset,
 	}
 
 out:
-	free(full_path);
 	return ret;
 }
 
@@ -569,11 +722,18 @@ static int process_clone(const char *path, u64 offset, u64 len,
 	struct btrfs_receive *r = user;
 	struct btrfs_ioctl_clone_range_args clone_args;
 	struct subvol_info *si = NULL;
-	char *full_path = path_cat(r->full_subvol_path, path);
+	char full_path[PATH_MAX];
 	char *subvol_path = NULL;
-	char *full_clone_path = NULL;
+	char full_clone_path[PATH_MAX];
 	int clone_fd = -1;
 
+	ret = path_cat_out(full_path, r->full_subvol_path, path);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: clone: source path invalid: %s\n",
+				path);
+		goto out;
+	}
+
 	ret = open_inode_for_write(r, full_path);
 	if (ret < 0)
 		goto out;
@@ -581,10 +741,10 @@ static int process_clone(const char *path, u64 offset, u64 len,
 	si = subvol_uuid_search(&r->sus, 0, clone_uuid, clone_ctransid, NULL,
 			subvol_search_by_received_uuid);
 	if (!si) {
-		if (memcmp(clone_uuid, r->cur_subvol->received_uuid,
+		if (memcmp(clone_uuid, r->cur_subvol.received_uuid,
 				BTRFS_UUID_SIZE) == 0) {
 			/* TODO check generation of extent */
-			subvol_path = strdup(r->cur_subvol->path);
+			subvol_path = strdup(r->cur_subvol_path);
 		} else {
 			ret = -ENOENT;
 			fprintf(stderr, "ERROR: did not find source subvol.\n");
@@ -609,9 +769,14 @@ static int process_clone(const char *path, u64 offset, u64 len,
 		subvol_path = strdup(si->path);
 	}
 
-	full_clone_path = path_cat3(r->root_path, subvol_path, clone_path);
+	ret = path_cat_out(full_clone_path, subvol_path, clone_path);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: clone: target path invalid: %s\n",
+				clone_path);
+		goto out;
+	}
 
-	clone_fd = open(full_clone_path, O_RDONLY | O_NOATIME);
+	clone_fd = openat(r->mnt_fd, full_clone_path, O_RDONLY | O_NOATIME);
 	if (clone_fd < 0) {
 		ret = -errno;
 		fprintf(stderr, "ERROR: failed to open %s. %s\n",
@@ -636,8 +801,6 @@ out:
 		free(si->path);
 		free(si);
 	}
-	free(full_path);
-	free(full_clone_path);
 	free(subvol_path);
 	if (clone_fd != -1)
 		close(clone_fd);
@@ -650,7 +813,31 @@ static int process_set_xattr(const char *path, const char *name,
 {
 	int ret = 0;
 	struct btrfs_receive *r = user;
-	char *full_path = path_cat(r->full_subvol_path, path);
+	char full_path[PATH_MAX];
+
+	ret = path_cat_out(full_path, r->full_subvol_path, path);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: set_xattr: path invalid: %s\n", path);
+		goto out;
+	}
+
+	if (strcmp("security.capability", name) == 0) {
+		if (g_verbose >= 3)
+			fprintf(stderr, "set_xattr: cache capabilities\n");
+		if (r->cached_capabilities_len)
+			fprintf(stderr,
+			  "WARNING: capabilities set multiple times per file: %s\n",
+				full_path);
+		if (len > sizeof(r->cached_capabilities)) {
+			fprintf(stderr,
+			  "ERROR: capabilities encoded to %d bytes, buffer too small\n",
+				len);
+			ret = -E2BIG;
+			goto out;
+		}
+		r->cached_capabilities_len = len;
+		memcpy(r->cached_capabilities, data, len);
+	}
 
 	if (g_verbose >= 2) {
 		fprintf(stderr, "set_xattr %s - name=%s data_len=%d "
@@ -667,7 +854,6 @@ static int process_set_xattr(const char *path, const char *name,
 	}
 
 out:
-	free(full_path);
 	return ret;
 }
 
@@ -675,7 +861,14 @@ static int process_remove_xattr(const char *path, const char *name, void *user)
 {
 	int ret = 0;
 	struct btrfs_receive *r = user;
-	char *full_path = path_cat(r->full_subvol_path, path);
+	char full_path[PATH_MAX];
+
+	ret = path_cat_out(full_path, r->full_subvol_path, path);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: remove_xattr: path invalid: %s\n",
+				path);
+		goto out;
+	}
 
 	if (g_verbose >= 2) {
 		fprintf(stderr, "remove_xattr %s - name=%s\n",
@@ -691,7 +884,6 @@ static int process_remove_xattr(const char *path, const char *name, void *user)
 	}
 
 out:
-	free(full_path);
 	return ret;
 }
 
@@ -699,7 +891,13 @@ static int process_truncate(const char *path, u64 size, void *user)
 {
 	int ret = 0;
 	struct btrfs_receive *r = user;
-	char *full_path = path_cat(r->full_subvol_path, path);
+	char full_path[PATH_MAX];
+
+	ret = path_cat_out(full_path, r->full_subvol_path, path);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: truncate: path invalid: %s\n", path);
+		goto out;
+	}
 
 	if (g_verbose >= 2)
 		fprintf(stderr, "truncate %s size=%llu\n", path, size);
@@ -713,7 +911,6 @@ static int process_truncate(const char *path, u64 size, void *user)
 	}
 
 out:
-	free(full_path);
 	return ret;
 }
 
@@ -721,7 +918,13 @@ static int process_chmod(const char *path, u64 mode, void *user)
 {
 	int ret = 0;
 	struct btrfs_receive *r = user;
-	char *full_path = path_cat(r->full_subvol_path, path);
+	char full_path[PATH_MAX];
+
+	ret = path_cat_out(full_path, r->full_subvol_path, path);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: chmod: path invalid: %s\n", path);
+		goto out;
+	}
 
 	if (g_verbose >= 2)
 		fprintf(stderr, "chmod %s - mode=0%o\n", path, (int)mode);
@@ -735,7 +938,6 @@ static int process_chmod(const char *path, u64 mode, void *user)
 	}
 
 out:
-	free(full_path);
 	return ret;
 }
 
@@ -743,7 +945,13 @@ static int process_chown(const char *path, u64 uid, u64 gid, void *user)
 {
 	int ret = 0;
 	struct btrfs_receive *r = user;
-	char *full_path = path_cat(r->full_subvol_path, path);
+	char full_path[PATH_MAX];
+
+	ret = path_cat_out(full_path, r->full_subvol_path, path);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: chown: path invalid: %s\n", path);
+		goto out;
+	}
 
 	if (g_verbose >= 2)
 		fprintf(stderr, "chown %s - uid=%llu, gid=%llu\n", path,
@@ -757,8 +965,24 @@ static int process_chown(const char *path, u64 uid, u64 gid, void *user)
 		goto out;
 	}
 
+	if (r->cached_capabilities_len) {
+		if (g_verbose >= 2)
+			fprintf(stderr, "chown: restore capabilities\n");
+		ret = lsetxattr(full_path, "security.capability",
+				r->cached_capabilities,
+				r->cached_capabilities_len, 0);
+		memset(r->cached_capabilities, 0,
+				sizeof(r->cached_capabilities));
+		r->cached_capabilities_len = 0;
+		if (ret < 0) {
+			ret = -errno;
+			fprintf(stderr, "ERROR: restoring capabilities %s: %s\n",
+					path, strerror(-ret));
+			goto out;
+		}
+	}
+
 out:
-	free(full_path);
 	return ret;
 }
 
@@ -768,9 +992,15 @@ static int process_utimes(const char *path, struct timespec *at,
 {
 	int ret = 0;
 	struct btrfs_receive *r = user;
-	char *full_path = path_cat(r->full_subvol_path, path);
+	char full_path[PATH_MAX];
 	struct timespec tv[2];
 
+	ret = path_cat_out(full_path, r->full_subvol_path, path);
+	if (ret < 0) {
+		fprintf(stderr, "ERROR: utimes: path invalid: %s\n", path);
+		goto out;
+	}
+
 	if (g_verbose >= 2)
 		fprintf(stderr, "utimes %s\n", path);
 
@@ -785,10 +1015,23 @@ static int process_utimes(const char *path, struct timespec *at,
 	}
 
 out:
-	free(full_path);
 	return ret;
 }
 
+static int process_update_extent(const char *path, u64 offset, u64 len,
+		void *user)
+{
+	if (g_verbose >= 2)
+		fprintf(stderr, "update_extent %s: offset=%llu, len=%llu\n",
+				path, (unsigned long long)offset,
+				(unsigned long long)len);
+
+	/*
+	 * Sent with BTRFS_SEND_FLAG_NO_FILE_DATA, nothing to do.
+	 */
+
+	return 0;
+}
 
 static struct btrfs_send_ops send_ops = {
 	.subvol = process_subvol,
@@ -811,13 +1054,16 @@ static struct btrfs_send_ops send_ops = {
 	.chmod = process_chmod,
 	.chown = process_chown,
 	.utimes = process_utimes,
+	.update_extent = process_update_extent,
 };
 
-static int do_receive(struct btrfs_receive *r, const char *tomnt, int r_fd,
-		      u64 max_errors)
+static int do_receive(struct btrfs_receive *r, const char *tomnt,
+		      char *realmnt, int r_fd, u64 max_errors)
 {
+	u64 subvol_id;
 	int ret;
 	char *dest_dir_full_path;
+	char root_subvol_path[PATH_MAX];
 	int end = 0;
 
 	dest_dir_full_path = realpath(tomnt, NULL);
@@ -836,20 +1082,24 @@ static int do_receive(struct btrfs_receive *r, const char *tomnt, int r_fd,
 		goto out;
 	}
 
-	ret = find_mount_root(dest_dir_full_path, &r->root_path);
-	if (ret < 0) {
-		fprintf(stderr,
-			"ERROR: failed to determine mount point for %s: %s\n",
-			dest_dir_full_path, strerror(-ret));
-		ret = -EINVAL;
-		goto out;
-	}
-	if (ret > 0) {
-		fprintf(stderr,
+	if (realmnt[0]) {
+		r->root_path = realmnt;
+	} else {
+		ret = find_mount_root(dest_dir_full_path, &r->root_path);
+		if (ret < 0) {
+			fprintf(stderr,
+				"ERROR: failed to determine mount point for %s: %s\n",
+				dest_dir_full_path, strerror(-ret));
+			ret = -EINVAL;
+			goto out;
+		}
+		if (ret > 0) {
+			fprintf(stderr,
 			"ERROR: %s doesn't belong to btrfs mount point\n",
 			dest_dir_full_path);
-		ret = -EINVAL;
-		goto out;
+			ret = -EINVAL;
+			goto out;
+		}
 	}
 	r->mnt_fd = open(r->root_path, O_RDONLY | O_NOATIME);
 	if (r->mnt_fd < 0) {
@@ -859,6 +1109,33 @@ static int do_receive(struct btrfs_receive *r, const char *tomnt, int r_fd,
 		goto out;
 	}
 
+	/*
+	 * If we use -m or a default subvol we want to resolve the path to the
+	 * subvolume we're sitting in so that we can adjust the paths of any
+	 * subvols we want to receive in.
+	 */
+	ret = btrfs_list_get_path_rootid(r->mnt_fd, &subvol_id);
+	if (ret) {
+		fprintf(stderr, "ERROR: couldn't resolve our subvolid %d\n",
+			ret);
+		goto out;
+	}
+
+	root_subvol_path[0] = 0;
+	ret = btrfs_subvolid_resolve(r->mnt_fd, root_subvol_path,
+				     PATH_MAX, subvol_id);
+	if (ret) {
+		fprintf(stderr, "ERROR: couldn't resolve our subvol path\n");
+		goto out;
+	}
+
+	/*
+	 * Ok we're inside of a subvol off of the root subvol, we need to
+	 * actually set full_root_path.
+	 */
+	if (strlen(root_subvol_path))
+		r->full_root_path = root_subvol_path;
+
 	if (r->dest_dir_chroot) {
 		if (chroot(dest_dir_full_path)) {
 			ret = -errno;
@@ -894,6 +1171,14 @@ static int do_receive(struct btrfs_receive *r, const char *tomnt, int r_fd,
 		goto out;
 
 	while (!end) {
+		if (r->cached_capabilities_len) {
+			if (g_verbose >= 3)
+				fprintf(stderr, "clear cached capabilities\n");
+			memset(r->cached_capabilities, 0,
+					sizeof(r->cached_capabilities));
+			r->cached_capabilities_len = 0;
+		}
+
 		ret = btrfs_read_and_process_send_stream(r_fd, &send_ops, r,
 							 r->honor_end_cmd,
 							 max_errors);
@@ -916,17 +1201,8 @@ out:
 	}
 	free(r->root_path);
 	r->root_path = NULL;
-	free(r->write_path);
-	r->write_path = NULL;
-	free(r->full_subvol_path);
-	r->full_subvol_path = NULL;
 	r->dest_dir_path = NULL;
 	free(dest_dir_full_path);
-	if (r->cur_subvol) {
-		free(r->cur_subvol->path);
-		free(r->cur_subvol);
-		r->cur_subvol = NULL;
-	}
 	subvol_uuid_search_finit(&r->sus);
 	if (r->mnt_fd != -1) {
 		close(r->mnt_fd);
@@ -936,23 +1212,27 @@ out:
 		close(r->dest_dir_fd);
 		r->dest_dir_fd = -1;
 	}
+
 	return ret;
 }
 
 int cmd_receive(int argc, char **argv)
 {
 	char *tomnt = NULL;
-	char *fromfile = NULL;
+	char fromfile[PATH_MAX];
+	char realmnt[PATH_MAX];
 	struct btrfs_receive r;
 	int receive_fd = fileno(stdin);
 	u64 max_errors = 1;
-	int ret;
+	int ret = 0;
 
 	memset(&r, 0, sizeof(r));
 	r.mnt_fd = -1;
 	r.write_fd = -1;
 	r.dest_dir_fd = -1;
 	r.dest_dir_chroot = 0;
+	realmnt[0] = 0;
+	fromfile[0] = 0;
 
 	while (1) {
 		int c;
@@ -962,7 +1242,7 @@ int cmd_receive(int argc, char **argv)
 			{ NULL, 0, NULL, 0 }
 		};
 
-		c = getopt_long(argc, argv, "Cevf:", long_opts, NULL);
+		c = getopt_long(argc, argv, "Cevf:m:", long_opts, NULL);
 		if (c < 0)
 			break;
 
@@ -971,7 +1251,13 @@ int cmd_receive(int argc, char **argv)
 			g_verbose++;
 			break;
 		case 'f':
-			fromfile = optarg;
+			if (arg_copy_path(fromfile, optarg, sizeof(fromfile))) {
+				fprintf(stderr,
+				    "ERROR: input file path too long (%zu)\n",
+				    strlen(optarg));
+				ret = 1;
+				goto out;
+			}
 			break;
 		case 'e':
 			r.honor_end_cmd = 1;
@@ -982,6 +1268,15 @@ int cmd_receive(int argc, char **argv)
 		case 'E':
 			max_errors = arg_strtou64(optarg);
 			break;
+		case 'm':
+			if (arg_copy_path(realmnt, optarg, sizeof(realmnt))) {
+				fprintf(stderr,
+				    "ERROR: mount point path too long (%zu)\n",
+				    strlen(optarg));
+				ret = 1;
+				goto out;
+			}
+			break;
 		case '?':
 		default:
 			fprintf(stderr, "ERROR: receive args invalid.\n");
@@ -994,15 +1289,17 @@ int cmd_receive(int argc, char **argv)
 
 	tomnt = argv[optind];
 
-	if (fromfile) {
+	if (fromfile[0]) {
 		receive_fd = open(fromfile, O_RDONLY | O_NOATIME);
 		if (receive_fd < 0) {
 			fprintf(stderr, "ERROR: failed to open %s\n", fromfile);
-			return 1;
+			goto out;
 		}
 	}
 
-	ret = do_receive(&r, tomnt, receive_fd, max_errors);
+	ret = do_receive(&r, tomnt, realmnt, receive_fd, max_errors);
+
+out:
 
 	return !!ret;
 }
@@ -1032,5 +1329,8 @@ const char * const cmd_receive_usage[] = {
 	"--max-errors <N> Terminate as soon as N errors happened while",
 	"                 processing commands from the send stream.",
 	"                 Default value is 1. A value of 0 means no limit.",
+	"-m <mountpoint>  The root mount point of the destination fs.",
+	"                 If you do not have /proc use this to tell us where ",
+	"                 this file system is mounted.",
 	NULL
 };
